import pandas as pd
import numpy as np
import scipy.optimize as solver
import math
from .correlation_analysis import log_returns, excise
from .macro_analysis import annualized_std



# 投资组合波动性
def prtf_sd(weights, cov):
    return np.sqrt(np.dot(weights.T, np.dot(cov, weights)))


# 用蒙特卡洛方法模拟权重组合
def monte_carlos_stimulator(df):

    rts = log_returns(df)
    #rt_a = annualized_returns(rts)
    rt_a = rts.mean() * 252
    cov = rts.cov()
    cov_a = cov * 252

    # 生成n个随机投资组合，n为参数
    prtf_rt = []
    prtf_vol = []
    #prtf_w = []

    n_assets = len(df.columns)
    n_prtf = 100000


    for i in range(n_prtf):
        weights = np.random.random(n_assets)
        weights /= np.sum(weights)
        rt = np.dot(weights, rt_a)
        vol = prtf_sd(weights, cov_a)
        
        prtf_rt.append(rt)
        prtf_vol.append(vol)
        #prtf_w.append(weights)

    prtf = pd.DataFrame([prtf_rt, prtf_vol]).T    
    prtf.columns = ['returns', 'volatility']
    prtf['sharpe'] = prtf['returns'] / prtf['volatility']

    return prtf


# 最大夏普比率（负向）
def neg_sharpe(weights, returns, cov):
    std = prtf_sd(weights, cov)
    rt = np.dot(weights, returns)
    return -rt / std



# 用scipy solver（优化函数）找到最大夏普组合的波动性，收益率，权重
def max_sharpe_ratio(df, ytd_days=None, limits=None, date_len=None):

    if date_len != None:
        rts = excise(log_returns(df), date_len, ytd_days)
    else:
        rts = log_returns(df)

    rt_a = rts.mean() * 252
    cov = rts.cov()
    cov_a = cov * 252

    num_assets = len(rt_a)
    args = (rt_a, cov_a)
    x0 = np.array([1.0 / num_assets for x in range(num_assets)]) 
    constraints = ({'type': 'eq', 'fun': lambda x: np.sum(x) - 1})

    # 查看是否有自定义上下限
    if limits == None:
        bound = (0.0, 1.0)
        bounds = tuple(bound for asset in range(num_assets))
    else:
        bounds = tuple((asset['lower'], asset['upper']) for asset in limits)
    
    max_sharpe = solver.minimize(neg_sharpe, x0, args=args,
                        method='SLSQP', bounds=bounds, constraints=constraints)

    sd_max = prtf_sd(max_sharpe['x'], cov_a)
    rt_max = np.dot(max_sharpe['x'], rt_a)

    max_sharpe_allocation = pd.DataFrame(max_sharpe['x'], index=df.columns, columns=['allocation'])
    max_sharpe_allocation['allocation'] = [round(i*100, 2) for i in max_sharpe_allocation['allocation']]
    max_sharpe_allocation = max_sharpe_allocation.T

    return max_sharpe_allocation, sd_max, rt_max




# 用scipy solver（优化函数）找到最小波动性组合的波动性，收益率，权重
def min_variance(df, ytd_days=None, limits=None, date_len=None):

    if date_len != None:
        rts = excise(log_returns(df), date_len, ytd_days)
    else:
        rts = log_returns(df)

    rt_a = rts.mean() * 252
    cov = rts.cov()
    cov_a = cov * 252

    num_assets = len(rt_a)
    args = cov_a
    x0 = np.array([1.0 / num_assets for x in range(num_assets)]) 
    constraints = ({'type': 'eq', 'fun': lambda x: np.sum(x) - 1})

    # 查看是否有自定义上下限
    if limits == None:
        bound = (0.0, 1.0)
        bounds = tuple(bound for asset in range(num_assets))
    else:
        bounds = tuple((asset['lower'], asset['upper']) for asset in limits)

    min_vol = solver.minimize(prtf_sd, x0, args=args, method='SLSQP', bounds=bounds, constraints=constraints)

    sd_min = prtf_sd(min_vol['x'], cov_a)
    rt_min = np.dot(min_vol['x'], rt_a)

    min_vol_allocation = pd.DataFrame(min_vol['x'], index=df.columns, columns=['allocation'])
    min_vol_allocation['allocation'] = [round(i*100, 2) for i in min_vol_allocation['allocation']]
    min_vol_allocation = min_vol_allocation.T

    return min_vol_allocation, sd_min, rt_min


# 有效前沿收益
def efficient_return(returns, cov, target, limits=None):
    num_assets = len(returns)
    args = cov
    x0 = np.array([1.0 / num_assets for x in range(num_assets)]) 

    def port_returns(weight):
        return np.dot(weight, returns)

    constraints = (
                    {'type': 'eq', 'fun': lambda x: port_returns(x) - target},
                    {'type': 'eq', 'fun': lambda x: np.sum(x) - 1}
                    )
    
    # 查看是否有自定义上下限
    if limits == None:
        bound = (0.0, 1.0)
        bounds = tuple(bound for asset in range(num_assets))
    else:
        bounds = tuple((asset['lower'], asset['upper']) for asset in limits)

    result = solver.minimize(prtf_sd, x0=x0, args=args, method='SLSQP', bounds=bounds, constraints=constraints)
    
    sd = prtf_sd(result['x'], cov)
    rt = np.dot(result['x'], returns)

    ary = np.append(np.append(result['x'], sd), rt)
    idx = returns.index.to_list() + ['std', 'rt']

    return pd.Series(ary, index=idx)


# 有效前沿组合
def efficient_frontier(df, n_prtf, ytd_days=None, limits=None, date_len=None):

    if date_len != None:
        rts = excise(log_returns(df), date_len, ytd_days)
    else:
        rts = log_returns(df)

    rt_a = rts.mean() * 252
    cov = rts.cov()
    cov_a = cov * 252

    rt_min = min_variance(df)[2]
    rt_max = max_sharpe_ratio(df)[2]
    target = np.linspace(rt_min, rt_max, n_prtf)

    efficients = []
    for ret in target:
        efficients.append(efficient_return(rt_a, cov_a, ret, limits))
    
    return efficients


# 投资风险比
# 找出单个基金的收益风险比（年化）
def return_risks_ef(df, ytd_days=None, date_len=None):

    if date_len != None:
        rts = excise(log_returns(df), date_len, ytd_days)
    else:
        rts = log_returns(df)

    stds_annualized = annualized_std(rts)
    rt_annualized = rts.mean() * 252

    rt_risk = pd.DataFrame([stds_annualized, rt_annualized]).T
    rt_risk.columns = ['risks', 'returns']
    rt_risk['sharpe'] = round(rt_risk['returns'] / rt_risk['risks'], 2)

    return rt_risk


# 既定目标收益
def ef_target_returns(tgt_min, tgt_max, n_prtf, df, ytd_days=None, limits=None, date_len=None):

    if tgt_max < tgt_min:
        raise Exception("目标设定错误!")
    else:

        # 用最大夏普和最小方差组合定义目标收益上下限
        rt_min = min_variance(df)[2]
        rt_max = max_sharpe_ratio(df)[2]
        
        if tgt_min < rt_min:
            raise Exception("超过目标下限! 最低下限为" + str(math.ceil(rt_min * 1000) / 1000))
        elif tgt_max > rt_max:
            raise Exception("超过目标上限! 最高上限为" + str(math.floor(rt_max * 1000) / 1000))
        else:
            if date_len != None:
                rts = excise(log_returns(df), date_len, ytd_days)
            else:
                rts = log_returns(df)
                rt_a = rts.mean() * 252
                cov = rts.cov()
                cov_a = cov * 252

                target = np.linspace(tgt_min, tgt_max, n_prtf)

                efficients = []
                for ret in target:
                    efficients.append(efficient_return(rt_a, cov_a, ret, limits))

    return efficients


# 既定目标风险
def ef_target_risks(tgt_min, tgt_max, n_prtf, df, limits=None, date_len=None):

    if tgt_max < tgt_min:
        raise Exception("目标设定错误!")
    else:
        # 由于配置组合是由最小方差所取得，只能用收益推导方差，无法用方差推导收益
        # 用最大夏普和最小方差组合定义目标收益上下限
        std_min = min_variance(df)[1]
        std_max = max_sharpe_ratio(df)[1]

        if tgt_min < std_min:
            raise Exception("超过目标下限! 最低下限为" + str(math.ceil(std_min * 1000) / 1000))
        elif tgt_max > std_max:
            raise Exception("超过目标上限! 最高上限为" + str(math.floor(std_max * 1000) / 1000))
        else:
            efficients = efficient_frontier(df, n_prtf, limits, date_len)

            efficient_portfolios = [p for p in efficients if (p['std'] >= tgt_min) & (p['std'] <= tgt_max)]

    return efficient_portfolios





import pandas as pd
import numpy as np
from datetime import datetime
#from WindPy import *


# 相关性分析

# 整理数据
def clean_data(df):

    df = df.pivot(index='report_date', columns='fund_code', values='anv_y') # values后期要改成复权累计净值
    df.index = pd.to_datetime(df.index)

    return df

# 对数收益率
def log_returns(df):

    return np.log(df / df.shift(1))

# 累计收益率
def cumulative_returns(rt):
    return np.exp(rt.cumsum())


# 摘取函数
def excise(df, date_len, ytd_days):

    len_dict = {'5yr': 1260, 
                '3yr': 756,
                '1yr': 252,
                'ytd': ytd_days,
                '6mo': 126,
                '3mo': 63,
                '1mo': 21}


    # 摘取的长度必须要小于或等于原有的长度，否则新df跟原df一致
    if len_dict[date_len] <= len(df):
    
        df_new = df.iloc[-len_dict[date_len]:].copy()
    
    else:
        df_new = df.copy()

    return df_new



# 为收益趋势曲线图表准备开始日期一致的收益数据
def collect_rt_plot(df, ytd_days=None, date_len=None):    

    df = clean_data(df)
    
    # 找到收益数据一致有效的开始日期
    rt = log_returns(df)
    idx_start = rt.loc[:rt.dropna(how='any').index[0]].index[-2]
    rt = rt.loc[idx_start:].copy()
    rt.iloc[0] = np.zeros(len(rt.columns))

    # 摘取设定的时间跨度
    if date_len != None:
        rt = excise(rt, date_len, ytd_days)


    return rt

# 相关系数(时间跨度为参数)
def cal_corr(rt):

    # rt为收益，date_len为时间跨度
    # 返回相关性矩阵

    return rt.corr()









import pandas as pd
import numpy as np
from .correlation_analysis import log_returns, cumulative_returns, excise


# 收益风险图

# 年化波动率
def annualized_std(rt):

    return rt.std() * np.sqrt(252)

# 年化收益率
def annualized_returns(rt):

    cum_rt = cumulative_returns(rt)

    return pd.Series([cum_rt[i].dropna().iloc[-1] ** (252 / len(cum_rt[i].dropna())) - 1 for i in cum_rt.columns], index=cum_rt.columns)

# 收益风险组合
def return_risks(df, date_len=None):

    if date_len != None:
        rt = excise(log_returns(df), date_len)
    else:
        rt = log_returns(df)

    stds_annualized = round(annualized_std(rt) * 100, 2)
    rt_annualized = round(annualized_returns(rt) * 100, 2)

    rt_risk = pd.DataFrame([stds_annualized, rt_annualized]).T
    rt_risk.columns = ['risks', 'returns']
    rt_risk['sharpe'] = round(rt_risk['returns'] / rt_risk['risks'], 2)

    return rt_risk



# # coding: utf-8
# import codecs
# import os
# import sys

# # try:
# from setuptools import setup
# # except:
# #     from distutils.core import setup

# """
# 打包的用的setup必须引入，
# """

# #!/usr/bin/env python
# # -*- coding: utf-8 -*-
 

# import sys
# if sys.version_info < (2, 5):
#     sys.exit('Python 3.6 or greater is required.')
 
 
# with open('README.md', 'rb') as fp:
#     readme = fp.read()
 
# # 版本号，自己随便写
# VERSION = "1.0.0"

# LICENSE = "MIT"

 
# setup(
#     name='model_builder',
#     version=VERSION,
#     description=(
#         '在搭建模型时必要的基础函数'
#     ),
#     long_description=readme,
#     author='ivey',
#     author_email='iveyzhang@outlook.com',
#     license=LICENSE,
#     py_mudoles=['model_builder'],
#     packages=['model_builder'],
#     platforms=["all"],
#     url='None',
#     install_requires=[  
#         'pandas',  
#         'numpy' 
#         ],  
#     classifiers=[
#         'Development Status :: 1 - Beta',
#         'Operating System :: Windows',
#         'Intended Audience :: Developers',
#         'License :: OSI Approved :: BSD License',
#         'Programming Language :: Python',
#         'Programming Language :: Python :: Implementation',
#         'Programming Language :: Python :: 3.6',
#         'Programming Language :: Python :: 3.8',
#         'Topic :: Software Development :: Libraries'
#     ],
# )


# # URL 你这个包的项目地址，如果有，给一个吧，没有你直接填写在PyPI你这个包的地址也是可以的
# # INSTALL_REQUIRES 模块所依赖的python模块
# # 以上字段不需要都包含


from setuptools import setup, find_packages

setup(
    name='ivtest',
    version='0.0.2',
    description='my test',
    author='ivey',
    author_email='iveyzhang@outlook.com',
    py_modules=["ivtest"],
    packages=['ivtest'],
    license='MIT',
    platform='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: System :: Logging',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    install_requires=['pandas', 'numpy']
)